#!C:\Perl\bin\perl
#
# board.cgi - Simple Message Board.
#
# Copyright (C) 2006 by Hiroshi Yuki.
# <hyuki@hyuki.com>
# http://www.hyuki.com/
#
# This program is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
################
use strict;
use warnings;
use lib "./lib";
use CGI;
use CGI::Carp qw(fatalsToBrowser);
use Digest::MD5 qw(md5_hex);
use HTML::Template;
use Data::Dumper;
use Yuki::YukiWikiDB;

################
# Configuration.
# Your CGI url.
my $cgi_url = "http://www.example.com/board/board.cgi";
# Your secret key.
my $secret_key = "A5D927E1E346E3A4C9E7481AE1113A41";
# Valid user list.
my %user_to_password = (
    # 'user_name' => 'password'
    'hyuki' => 'perl',
    'rubyco' => 'ruby',
    'guest' => 'guest',
);
# WARNING: FOR SECURITY DEMONSTRATION!
my $SKIP_VERIFY_WRITER = 0;
################
# Globals.
my $DEBUG = 0;
my $debug_log = "";
my $expire_seconds = 3600;
my $cgi = new CGI;
my %context = (
    TITLE => 'メッセージボード',
    CSS_URL => 'style.css',
    CGI_URL => $cgi_url,
    CHARSET => 'UTF-8',
    AUTHENTICATED => 0,
    USER_NAME => '',
    LOGOUT_URL => "${cgi_url}?mycmd=logout",
    SESSION => '',
    MAX_MSG => 5,
    MAX_MSG_LENGTH => 60,
    COOKIE => '',
    MSG_LIST => [],
    SKIP_VERIFY_WRITER => $SKIP_VERIFY_WRITER,
);
$cgi->charset($context{CHARSET});
my %template_settings = (
    die_on_bad_params => 0,
    global_vars => 1,
);
my $user_db_name = 'user';
my $msg_db_name = 'msg';
my %user_db;
my %msg_db;

sub main();
sub error_exit($);
sub debug_log($);
sub create_new_session_for_user_name($);
sub get_user_name_by_session($);
sub create_cookie_for_session($);
sub create_unique_id($);
sub do_login();
sub do_logout();
sub do_write_msg();
sub do_delete_msgs();
sub do_display_msgs();
sub read_msg_list();
sub verify_msg($);
sub convert_and_escape_msg($);
sub write_msg($);
sub delete_msgs();
sub get_now();
sub open_db();
sub close_db();
sub verify_writer();
sub is_valid_user($$);
sub delete_user_cookie();
sub setup_context($$);

main();
exit;

# Main.
sub main() {
    # Setup context by cookie.
    my $session = $cgi->cookie('mysession');
    my $user_name = get_user_name_by_session($session);
    if (defined $user_name) {
        setup_context($user_name, $session);
    }

    # Dispatch commands.
    my $mycmd = $cgi->param('mycmd') || '';
    if ($mycmd eq 'login') {
        do_login();
    } elsif ($mycmd eq 'logout') {
        do_logout();
    } elsif ($mycmd eq 'write') {
        do_write_msg();
    } elsif ($mycmd eq 'delete') {
        do_delete_msgs();
    } else {
        do_display_msgs();
    }

    # Show debug log.
    if ($DEBUG) {
        error_exit('Done');
    }
}

# Show error message and exit.
sub error_exit($) {
    my $msg = shift;
    print $cgi->header(
        -expires => 'now', 
        -charset => $context{CHARSET},
    );
    print $cgi->start_html('Error');
    print $cgi->h1('Error');
    print $cgi->start_pre;
    print $cgi->escapeHTML($msg);
    print $cgi->end_pre;
    if ($DEBUG) {
        print $cgi->h1('debug_log');
        print $cgi->start_pre;
        print $cgi->escapeHTML($debug_log);
        print $cgi->end_pre;
    }
    print $cgi->end_html;
    exit;
}

# Append a log entry.
sub debug_log($) {
    my $s = shift;
    if ($DEBUG) {
        $debug_log .= "$s\n";
    }
}

# Create new session for the user.
sub create_new_session_for_user_name($) {
    my ($user_name) = @_;

    open_db();

    # Check alive session.
    my %alive_session;
    while (my ($key, $line) = each %user_db) {
        if (not $line =~ /^(.+?),(.+?),(.+?)$/) {
            debug_log("create_new_session_for_user_name: delete: " . Dumper(\$line));
            delete $user_db{$key};
            next;
        }
        my ($session, $expire, $name) = ($1, $2, $3);

        # Delete the user.
        if ($name eq $user_name) {
            debug_log("create_new_session_for_user_name: delete: " . Dumper(\$line));
            delete $user_db{$key};
            next;
        }

        # Delete expired user.
        if (time < $expire) {
            debug_log("create_new_session_for_user_name: expired: " . Dumper(\$line));
            delete $user_db{$key};
            next;
        }

        # Memorize the session.
        $alive_session{$session}++;
    }

    # Create a new session.
    my $new_session;
    my $trial = 0;
    while (1) {
        $new_session = create_unique_id($trial);
        if (not $alive_session{$new_session}) {
            last;
        }
        $trial++;
    }

    # Construct a new user info.
    my $expire = time + $expire_seconds;
    my $user_info = "$new_session,$expire,$user_name";
    $user_db{$new_session} = $user_info;

    debug_log("create_new_session_for_user_name: new_session = $new_session, user_info = $user_info");

    close_db();

    return $new_session;
}

# Load user info.
sub get_user_name_by_session($) {
    my $session = shift;
    return if not $session;
    return if not $session =~ /^[0-9a-f]+$/;

    open_db();
    my $line = $user_db{$session};
    close_db();

    debug_log("get_user_name_by_session: session = $session");
    debug_log("get_user_name_by_session: line = $line");

    if (not $line =~ /^(.+?),(.+?),(.+?)$/) {
        debug_log("get_user_name_by_session: invalid line = $line.");
        open_db();
        delete $user_db{$session};
        close_db();
        return;
    }

    my ($expire, $user_name) = ($2, $3);
    if ($expire < time) {
        debug_log("get_user_name_by_session: session = $session is already expired.");
        open_db();
        delete $user_db{$session};
        close_db();
        return;
    }
    
    return $user_name;
}

# Create a cookie.
sub create_cookie_for_session($) {
    my $session = shift;
    my $cookie = $cgi->cookie(
        -name => 'mysession',
        -value => $session,
        -expires => "+" . $expire_seconds . "s",
    );
    return "Set-Cookie: $cookie";
}

# Create unique_id for a new session.
sub create_unique_id($) {
    my $trial = shift;
    return md5_hex($secret_key . localtime() . $$ . rand() . $trial);
}

# Write a message.
sub do_write_msg() {
    verify_writer();
    my $mymsg = $cgi->param("mymsg");
    if (length($mymsg) > 0) {
        verify_msg($mymsg);
        write_msg(convert_and_escape_msg($mymsg));
    }
    do_display_msgs();
}

# Delete messages.
sub do_delete_msgs() {
    verify_writer();
    delete_msgs();
    do_display_msgs();
}

# Login.
sub do_login() {
    my $user_name = $cgi->param('myname');
    my $password = $cgi->param('mypassword');

    if (not is_valid_user($user_name, $password)) {
        error_exit('Invalid user');
    }

    my $new_session = create_new_session_for_user_name($user_name);
    setup_context($user_name, $new_session);

    my $template = HTML::Template->new(
        filename => "do_login.txt",
        %template_settings,
    );
    $template->param(%context);
    print $template->output;
}

# Logout.
sub do_logout() {
    delete_user_cookie;

    my $template = HTML::Template->new(
        filename => "do_logout.txt",
        %template_settings,
    );

    if ($context{USER_NAME}) {
        open_db();
        # Delete the user.
        while (my ($key, $line) = each %user_db) {
            if (not $line =~ /^(.+?),(.+?),(.+?)$/) {
                debug_log("do_logout: invalid line = $line.");
                delete $user_db{$key};
                next;
            }
            my ($session, $expire, $user_name) = ($1, $2, $3);
            if ($user_name eq $context{USER_NAME}) {
                debug_log("do_logout: delete = $key");
                delete $user_db{$key};
            }
        }
        close_db();
    }

    $template->param(%context);
    print $template->output;
}

# Display messages.
sub do_display_msgs() {
    my $template = HTML::Template->new(
        filename => "do_display_msgs.txt",
        %template_settings,
    );

    $context{MSG_LIST} = read_msg_list();

    $template->param(%context);
    print $template->output;
}

sub read_msg_list() {
    open_db();
    my @msg_list;
    while (my ($key, $value) = each %msg_db) {
        next if $key !~ /^\d+$/;

        my %msg;
        @msg{qw(
            WRITER_NAME
            DATE
            MSG_ID
            MSG_BODY_ESCAPED
        )} = split(/,/, $value);

        # Extra.
        if ($msg{WRITER_NAME} eq $context{USER_NAME}) {
            $msg{WRITER_AUTHENTICATED} = 1;
        } else {
            $msg{WRITER_AUTHENTICATED} = 0;
        }

        debug_log("read_msg_list: push: " . Dumper(\%msg));

        push(@msg_list, \%msg);
    }
    # Sort by date.
    my @rest = sort { $b->{DATE} cmp $a->{DATE} } @msg_list;
    # Delete old messages.
    @msg_list = splice(@rest, 0, $context{MAX_MSG});
    for my $h (@rest) {
        debug_log("read_msg_list: delete: $h->{MSG_ID}");
        delete $msg_db{$h->{MSG_ID}};
    }
    close_db();
    return \@msg_list;
}

sub convert_and_escape_msg($) {
    my $msg = shift;
    
    # Escape.
    my $escaped_msg = $cgi->escapeHTML($msg);
    $escaped_msg =~ s/,/&#44;/g;

    return $escaped_msg;
}

sub verify_msg($) {
    my $msg = shift;
    # Check length.
    if ($context{MAX_MSG_LENGTH} < length($msg)) {
        error_exit("Too long message.");
    }
}

sub write_msg($) {
    my $escaped_msg = shift;

    open_db();

    if ($msg_db{LAST_MSG_ID}) {
        $msg_db{LAST_MSG_ID} = $msg_db{LAST_MSG_ID} + 1;
    } else {
        $msg_db{LAST_MSG_ID} = 1;
    }

    my @msg;
    push(@msg, $context{USER_NAME});
    push(@msg, get_now());
    push(@msg, $msg_db{LAST_MSG_ID});
    push(@msg, $escaped_msg);

    my $msg = join(',', @msg);

    $msg_db{$msg_db{LAST_MSG_ID}} = $msg;

    close_db();
}

sub delete_msgs() {
    open_db();

    my %mymsgid = map { $_ => 1 } $cgi->param('mymsgid');
    for my $msgid (keys %mymsgid) {
        my $value = $msg_db{$msgid} or next;
        my %msg;
        @msg{qw(
            WRITER_NAME
            DATE
            MSG_ID
            MSG_BODY_ESCAPED
        )} = split(/,/, $value);
        if ($msg{MSG_ID} eq $msgid
        and $msg{WRITER_NAME} eq $context{USER_NAME}) {
            debug_log("delete_msgs: " . Dumper(\%msg));
            delete $msg_db{$msgid};
        }
    }

    close_db();
}

sub get_now() {
    my @week = qw(Sun Mon Tue Wed Thu Fri Sat);
    my ($sec, $min, $hour, $day, $mon, $year, $weekday) = localtime(time);
    return sprintf('%04d-%02d-%02d (%s) %02d:%02d:%02d',
        $year + 1900, $mon + 1, $day,
        $week[$weekday],
        $hour, $min, $sec);
}

sub open_db() {
    tie(%user_db, "Yuki::YukiWikiDB", $user_db_name) or error_exit("(tie) $user_db_name");
    tie(%msg_db, "Yuki::YukiWikiDB", $msg_db_name) or error_exit("(tie) $msg_db_name");
}

sub close_db() {
    untie(%msg_db);
    untie(%user_db);
}

sub verify_writer() {
    # Check request method.
    if ($cgi->request_method ne 'POST') {
        error_exit("Invalid request method.");
    }

    # Check if the writer is authenticated.
    if (not $context{AUTHENTICATED}) {
        error_exit("Not authenticated.");
    }

    # WARNING: FOR SECURITY DEMONSTRATION!
    if ($SKIP_VERIFY_WRITER) {
        return;
    }

    # Check if the writer came from this form.
    if ($cgi->cookie('mysession') ne $cgi->param('mysession')) {
        error_exit("Invalid session.");
    }
}

sub is_valid_user($$) {
    my ($name, $password) = @_;
    if (defined($user_to_password{$name}) and $user_to_password{$name} eq $password) {
        return 1;
    } else {
        return 0;
    }
}

sub setup_context($$) {
    my ($user_name, $session) = @_;
    $context{USER_NAME} = $user_name;
    $context{AUTHENTICATED} = 1;
    $context{SESSION} = $session;
    $context{COOKIE} = create_cookie_for_session($session);
}

sub delete_user_cookie() {
    my $cookie = $cgi->cookie(
        -name => 'mysession',
        -value => '',
        -expires => 'now',
    );
    $context{COOKIE} = "Set-Cookie: $cookie";
}
__END__
